
Require Import Tactics.
Require Import Sequence.
Require Import Dynamic.
Require Import Hwf.
Require Import Relation.
Require Import Match.
Require Import History.
Require Import Truncate.
Require Import Acyclic.
Require Import Decide.



Definition allexec H :=
  forall i p, H {{ ev_init i p }} -> exec H i.


Definition consistent H :=
  exists H', trco (H;; H') /\ acyclic (co (H;; H')) /\ allexec (H;; H') /\ ~ H' {{ ev_is i a | i a }}.


(* from-read *)
Definition fr H i i' :=
  exists i'',
    rf H i'' i
    /\ cop H i'' i'
    /\ i <> i'.


(* sequentially consistent order *)
Definition sc R H i i' :=
  R i i' \/ co H i i' \/ rf H i i' \/ fr H i i'.


(* reduced sc = sc without R *)
Definition rsc H i i' :=
  co H i i' \/ rf H i i' \/ fr H i i'.


(* pushed program order, for theorem 1 *)
Definition ppo H i i' :=
  exists is,
    isp H is
    /\ vop H i is
    /\ xop H is i'.


(* strong pushed program order, for theorem 3 *)
Definition sppo H i i' :=
  exists is,
    isp H is
    /\ H {{ ev_fence (fc_post at_x) is }}
    /\ vop H i is
    /\ xop H is i'.


(* synchronized before *)
Definition sb H i i' :=
  exists i'',
    vop H i i''
    /\ H {{ ev_fence (fc_post at_x) i'' }}
    /\ poq H i'' i'.


Definition fpo F H i i' :=
  F i /\ po H i i'.


Inductive conflict : history -> identifier -> identifier -> Prop :=
| conflict_ww {H i i' l}
    : writesto H i l
      -> writesto H i' l
      -> conflict H i i'

| conflict_wr {H i i' l}
    : writesto H i l
      -> reads H i' l
      -> conflict H i i'

| conflict_rw {H i i' l}
    : reads H i l
      -> writesto H i' l
      -> conflict H i i'
.


Hint Constructors conflict : seqcon.


Definition accesses H i l :=
  reads H i l \/ writesto H i l.


(* data-race-free, for members of F *)
Definition racefree (F : identifier -> Prop) H : Prop :=
  forall i i' p p',
    H {{ ev_init i p }}
    -> H {{ ev_init i' p' }}
    -> p <> p'
    -> conflict H i i'
    -> F i
    -> sb H i i' \/ sb H i' i.


Definition sc1 H := sc (ppo H) H.
Definition sc2 H := sc (po H) H.
Definition sc3 F H := sc (union (sppo H) (fpo F H)) H.


(* Service lemmas *)

Lemma allexec_inited :
  forall H i,
    allexec H
    -> inited H i
    -> exec H i.
Proof.
intros H i Hallexec (p & Hinited).
eapply Hallexec; eauto.
Qed.


Lemma allexec_is :
  forall H i a,
    trco H
    -> allexec H
    -> H {{ ev_is i a }}
    -> exec H i.
Proof.
intros H i a Htrco Hallexec His.
apply allexec_inited; auto; [].
eapply is_inited; eauto; done.
Qed.


Lemma allexec_po_fst :
  forall H i i',
    allexec H
    -> po H i i'
    -> exec H i.
Proof.
intros H i i' Hallexec Hpo.
apply allexec_inited; auto; [].
exact (po_inited _#3 Hpo andel).
Qed.


Lemma allexec_po_snd :
  forall H i i',
    allexec H
    -> po H i i'
    -> exec H i'.
Proof.
intros H i i' Hallexec Hpo.
apply allexec_inited; auto; [].
exact (po_inited _#3 Hpo ander).
Qed.


Lemma fr_isr_isw :
  forall H i i',
    trco H
    -> fr H i i'
    -> isr H i /\ isw H i'.
Proof.
intros H i i' Htrco Hfr.
destruct Hfr as (i'' & Hrf & Hcop & _).
split.
  pose proof (rf_wf_prop _#3 Htrco Hrf) as (l & _ & Hinr).
  eexists; eauto.

  pose proof (plus_plusr _#4 Hcop) as (i''' & _ & Hco).
  pose proof (co_isw _#3 Htrco Hco) as (_ & Hisw).
  assumption.
Qed.


Lemma fr_inited :
  forall H i i',
    trco H
    -> fr H i i'
    -> inited H i /\ inited H i'.
Proof.
intros H i i' Htrco Hfr.
destruct Hfr as (iw & Hrf & Hcop & _).
split.
  eapply rf_inited; eauto; done.

  eapply cop_inited; eauto; done.
Qed.


Lemma fr_itemizable_right :
  forall H i,
    trco H
    -> itemizable (fun i' => fr H i i').
Proof.
intros H i Htrco.
#2 apply (itemizable_compose_cdr _ _ (fun is => rf H is i) (fun is i' => cop H is i' /\ i <> i')).
  apply rf_itemizable_left; done.

  intros is _.
  #2 apply itemizable_and.
    apply cop_right_itemizable; auto; done.

    intro i'.
    apply dec_not; [].
    apply (eq_dec_decidable _ eq_ident_dec); auto; done.
Qed.


Lemma sppo_ppo :
  forall H i i', sppo H i i' -> ppo H i i'.
Proof.
intros H i i' Hsppo.
destruct Hsppo as (? & ? & _ & ?).
unfold ppo; eauto.
Qed.


Lemma rsc_inited :
  forall H i i',
    trco H
    -> rsc H i i'
    -> inited H i /\ inited H i'.
Proof.
intros H i i' Htrco Hrsc.
destruct Hrsc as [|[|]]; auto using co_inited, rf_inited, fr_inited.
Qed.


Lemma rscs_scs :
  forall R H i i',
    star (rsc H) i i' -> star (sc R H) i i'.
Proof.
intros R H i i' Hrscs.
apply (star_mono _ (rsc H)); auto; [].
intros; right; assumption.
Qed.


Lemma reads_accesses :
  forall H i l,
    reads H i l -> accesses H i l.
Proof.
intros; left; assumption.
Qed.


Lemma writesto_accesses :
  forall H i l,
    writesto H i l -> accesses H i l.
Proof.
intros; right; assumption.
Qed.


Lemma ppo_app_incl :
  forall H H' i1 i2,
    ppo H i1 i2
    -> ppo (H;; H') i1 i2.
Proof.
intros H H' i1 i2 Hppo.
destruct Hppo as (is & Hisp & Hvop & Hxop).
exists is.
#3 repeat2 split.
  apply in_or_app; right; assumption.

  apply vop_app_incl; assumption.

  apply xop_app_incl; assumption.
Qed.


Lemma sppo_app_incl :
  forall H H' i1 i2,
    sppo H i1 i2
    -> sppo (H;; H') i1 i2.
Proof.
intros H H' i1 i2 Hppo.
destruct Hppo as (is & Hisp & Hfence & Hvop & Hxop).
exists is.
#4 repeat2 split.
  apply in_or_app; right; assumption.

  apply in_or_app; right; assumption.

  apply vop_app_incl; assumption.

  apply xop_app_incl; assumption.
Qed.


Lemma sb_cons_incl :
  forall H h i i',
    sb H i i'
    -> sb (H; h) i i'.
Proof.
intros H h i i' Hsb.
destruct Hsb as (i'' & Hvop & Hfence & Hpoq).
exists i''.
do2 2 split.
  apply vop_cons_incl; auto; done.
  
  right; assumption.

  apply poq_cons_incl; auto; done.
Qed.


Lemma sb_app_incl :
  forall H H' i i',
    sb H i i'
    -> sb (H;; H') i i'.
Proof.
exact (cons_incl_imp_app_incl2 _#3 sb_cons_incl).
Qed.


Lemma fpo_app_incl :
  forall F H H' i1 i2,
    fpo F H i1 i2
    -> fpo F (H;; H') i1 i2.
Proof.
intros F H H' i1 i2 Hfpo.
destruct Hfpo as (Hf & Hpo).
split; auto; [].
apply po_app_incl; auto; done.
Qed.


Lemma conflict_is :
  forall H i i',
    conflict H i i'
    -> H {{ ev_is i a | a }} /\ H {{ ev_is i' a | a }}.
Proof.
intros H i i' Hconflict.
invert Hconflict;
intros;
repeat (match goal with
        | HH : writesto _ _ _ |- _ => destruct HH as (? & HH); invert HH
        | HH : reads _ _ _ |- _ => invert HH; clear HH
        end); eauto.
Qed.


Lemma conflict_truncate :
  forall H H' i1 i2,
    ~ H' {{ ev_is i a | i a }}
    -> conflict (H;; H') i1 i2
    -> conflict H i1 i2.
Proof.
intros H H' i1 i2 Hnis Hconflict.
invert Hconflict; intros; eauto using reads_truncate, writesto_truncate with seqcon.
Qed.


Lemma accesses_truncate :
  forall H H' i l,
    ~ H' {{ ev_is i' a | i' a }}
    -> accesses (H;; H') i l
    -> accesses H i l.
Proof.
intros H H' i l Hnis Haccess.
destruct Haccess.
  left; eapply reads_truncate; eauto; done.

  right; eapply writesto_truncate; eauto; done.
Qed.


Lemma ppo_decidable :
  forall H i i',
    trco H
    -> decidable (ppo H i i').
Proof.
intros H i i' Htrco.
apply itemizable_decidable_exists; [].
#3 do2 2 (apply itemizable_and_right).
  intro is.
  apply present_decidable; done.

  intro is.
  apply vop_decidable; auto; done.

  apply xop_itemizable_left; auto; done.
Qed.


Lemma sppo_decidable :
  forall H i i',
    trco H
    -> decidable (sppo H i i').
Proof.
intros H i i' Htrco.
apply itemizable_decidable_exists; [].
#4 do2 3 (apply itemizable_and_right).
  intro is.
  apply present_decidable; done.

  intro is.
  apply present_decidable; done.

  intro is.
  apply vop_decidable; auto; done.

  apply xop_itemizable_left; auto; done.
Qed.


(* First theorem: sequential consistency using ppo. *)

(* First we establish that an sc cycle contains an R edge. *)

Lemma fr_isw_cop :
  forall H i i',
    trco H
    -> fr H i i'
    -> isw H i
    -> cop H i i'.
Proof.
intros H i i' Htrco Hfr Hisw.
destruct Hfr as (i'' & Hrf & Hcop & Hneq).
apply star_neq_plus; auto; [].
eapply rwf_cop_cos; eauto; [].
apply rf_isw_rwf; auto; done.
Qed.


Lemma fr_cop_trans :
  forall H i1 i2 i3,
    trco H
    -> acyclic (co H)
    -> fr H i1 i2
    -> cop H i2 i3
    -> fr H i1 i3.
Proof.
intros H i1 i2 i3 Htrco Hcoacy Hfr Hcop23.
destruct Hfr as (iw & Hrf & Hcop & _).
exists iw.
#2 do2 2 split; auto.
  eapply plus_trans; eauto; done.

  intro; subst i3.
  destruct Hcoacy; [].
  exists i1.
  eapply star_plus_trans; eauto; [].
  eapply rwf_cop_cos; eauto; [].
  apply rf_isw_rwf; auto; [].
  eapply cop_isw; eauto; done.
Qed.


Lemma rscp_to_write :
  forall H i i',
    trco H
    -> acyclic (co H)
    -> plus (rsc H) i i'
    -> isw H i'
    -> cop H i i' \/ fr H i i'.
Proof.
intros H i i' Htrco Hcoacy Hrscp Hisw.
pose proof (plus_plusi _#4 Hrscp) as Hrscp'; clear Hrscp.
revert Hisw.
elim Hrscp'; clear i i' Hrscp'.
(* refl *)
intros i i' Hrsc Hisw.
have (rsc H i i') as Hrsc.
have (isw H i') as Hisw.
toshow (cop H i i' \/ fr H i i').
destruct Hrsc as [Hco | [Hrf | Hfr]].
  left.
  apply plus_one.
  assumption.

  left.
  apply plus_one; [].
  apply rwf_co; auto; [].
  apply rf_isw_rwf; auto; done.

  right.
  assumption.

(* step *)
intros i1 i2 i3 Hrsc Hrscp IH Hisw.
thus (cop H i2 i3 \/ fr H i2 i3) as Hacc by IH; clear IH Hisw.
toshow (cop H i1 i3 \/ fr H i1 i3).
destruct Hrsc as [Hco | [Hrf | Hfr]].
  have (co H i1 i2) as Hco.
  destruct Hacc as [Hcop | Hfr].
    left.
    have (cop H i2 i3) as Hcop.
    eapply plus_step; eauto; done.

    have (fr H i2 i3) as Hfr.
    left.
    exists i2; split; auto; [].
    apply plus_star; [].
    apply fr_isw_cop; auto; [].
    eapply co_isw; eauto; done.

  have (rf H i1 i2) as Hrf.
  left.
  destruct Hacc as [Hcop | Hfr].
    have (cop H i2 i3) as Hcop.
    pose proof (cop_isw _#3 Htrco Hcop) as (Hisw & _).
    eapply plus_step; eauto; [].
    apply rwf_co; auto; [].
    apply rf_isw_rwf; auto; done.

    have (fr H i2 i3) as Hfr.
    destruct Hfr as (i & Hrf' & Hcop & _).
    have (rf H i i2) as Hrf'.
    have (cop H i i3) as Hcop.
    thus (i = i1) as Heq by rf_fun.
    subst i.
    assumption.

  have (fr H i1 i2) as Hfr.
  right.
  eapply fr_cop_trans; eauto; [].
  destruct Hacc as [Hcop | Hfr']; auto; [].
  apply fr_isw_cop; auto; [].
  eapply fr_isr_isw; eauto; done.
Qed.


Lemma rscs_to_write :
  forall H i i',
    trco H
    -> acyclic (co H)
    -> star (rsc H) i i'
    -> isw H i'
    -> cos H i i' \/ fr H i i'.
Proof.
intros H i i' Htrco Hcoacy Hrscs Hisw.
so (star_plus _#4 Hrscs) as [Heq | Hrscp].
  subst; left; apply star_refl; done.

  so (rscp_to_write _#3 Htrco Hcoacy Hrscp Hisw) as [Hcop | Hfr].
    left; apply plus_star; auto; done.

    right; assumption.
Qed.


Lemma rsc_acyclic :
  forall H,
    trco H
    -> acyclic (co H)
    -> acyclic (rsc H).
Proof.
intros H Htrco Hcoacy.
intro Hcycle.
assert (exists i, isw H i /\ plus (rsc H) i i) as Hcycle'.
  destruct Hcycle as (i & (i' & Hrsc & Hrscs)).
  destruct Hrsc as [Hco | [Hrf | Hfr]].
    exists i.
    split.
      exact (co_isw _#3 Htrco Hco andel).

      exists i'; split; auto.
      left; auto.

    exists i.
    split.
      exact (rf_isw_isr _#3 Htrco Hrf andel).

      exists i'; split; auto.
      right; left; auto.

    exists i'.
    split.
      exact (fr_isr_isw _#3 Htrco Hfr ander).

      apply plusr_plus.
      exists i; split; auto.
      right; right; auto.
have (exists i : identifier, isw H i /\ plus (rsc H) i i) as Hcycle'.
clear Hcycle.
destruct Hcycle' as (i & Hisw & Hrscp).
destruct (rscp_to_write _#3 Htrco Hcoacy Hrscp Hisw) as [Hcop | Hfr].
  have (cop H i i) as Hcop.
  destruct Hcoacy; [].
  exists i; auto.

  have (fr H i i) as Hfr.
  destruct Hfr as (_ & _ & _ & Hneq).
  destruct Hneq; auto; done.
Qed.


Lemma sc_cycle_r :
  forall R H i,
    trco H
    -> acyclic (co H)
    -> plus (sc R H) i i
    -> exists i1 i2,
         R i1 i2
         /\ star (sc R H) i2 i1.
Proof.
intros R H i Htrco Hcoacy Hcycle.
destruct Hcycle as (i' & Hsc & Hscs).
destruct Hsc as [HR | Hrsc].
  exists i, i'; splitall; auto; done.

  change (rsc H i i') in Hrsc.
  thus (plus (rsc H) i i') as Hacc by plus_one; clear Hrsc.
  remember i as i0 in Hacc.
  revert i0 Heqi0 Hacc.
  elim Hscs; clear i i' Hscs.
  (* refl *)
  intros i i0 Heqi0 Hacc.
  subst i0.
  destruct (rsc_acyclic _ Htrco Hcoacy).
  exists i; auto.

  (* step *)
  intros i1 i2 i3 H12 H23 IH i0 Heqi0 Hacc.
  subst i0.
  destruct H12 as [HR | Hrsc].
    exists i1, i2.
    split; auto; [].
    eapply star_trans; eauto; [].
    apply (star_mono _ (rsc H)); [|].
      intros; right; auto; done.

      apply plus_star; auto; done.

    have (co H i1 i2 \/ rf H i1 i2 \/ fr H i1 i2) as Hrsc.
    change (rsc H i1 i2) in Hrsc.
    apply (IH _ (eq_refl _)); [].
    eapply plus_trans; eauto using plus_one.
Qed.


(* Main lemma *)

(* If  ia -xo->+ i -rsc->* i' -vo->+ ib
   and exec ia
   and isp ib
   then
       ia -to-> ib
*)
Lemma rscs_then_push :
  forall H i i' ia ib,
    trco H
    -> acyclic (co H)
    -> exec H ia
    -> isp H ib
    -> xos H ia i
    -> star (rsc H) i i'
    -> vop H i' ib
    -> to H ia ib.
Proof.
intros H i2 i3 ib ia Htrco Hcoacy Hexecb Hisp Hxos2 Hrscs Hvop3.
so (swapp (exec_decidable H ia)) as [Hnexeca | Hexeca].
  apply to_from_nexec; auto; done.
so (star_plus _#4 Hrscs) as [Heq | Hrscp].
  subst i3.
  apply (toq_to_trans _ _ i2); auto; [|].
    apply xos_toq'; auto; done.

    apply vop_to; auto; done.
(* Assume, without loss of generality, that i3 is a write. *)
assert (exists i3', isw H i3' /\ star (rsc H) i2 i3' /\ vop H i3' ia) as Hprop.
  so (plus_plusr _#4 Hrscp) as (i4 & Hrscs24 & Hrsc43); clear Hrscp.
  destruct Hrsc43 as [Hco | [Hrf | Hfr]].
    exists i3.
    do2 2 split; auto; [].
    eapply co_isw; eauto; done.

    exists i4.
    do2 2 split; auto; [|].
      eapply rf_isw_isr; eauto; done.

      eapply star_plus_trans; eauto; [].
      apply star_one; [].
      right; left; assumption.

    exists i3.
    do2 2 split; auto; [].
    eapply fr_isr_isw; eauto; done.
clear i3 Hvop3 Hrscs Hrscp.
destruct Hprop as (i3 & Hisw3 & Hrscs & Hvop3).
cut (~ toq H ia ib).
  intro Hntoq.
  so (trace_trichot _#3 Htrco Hexeca Hexecb) as [|[|]].
    destruct Hntoq.
    right; assumption.

    destruct Hntoq.
    left; subst; reflexivity.

    assumption.
intro Htoq.
assert (vop H i3 ib) as Hvop3b.
  eapply plus_star_trans; eauto; [].
  destruct Htoq as [Heq | Hto].
    subst; apply star_refl; done.

    apply star_one; [].
    right; right.
    split; auto; done.
clear Hvop3 Htoq.
so (rscs_to_write _#3 Htrco Hcoacy Hrscs Hisw3) as [Hcos23 | Hfr].
  (* write case *)
  apply Hcoacy; auto; [].
  exists i3.
  eapply plus_star_trans; eauto; [].
  destruct Hisw3 as (l & Hwrite3).
  apply plus_one; [].
  apply co_write; [].
  #2 eapply wwp_vt; eauto.
    exists ib; auto; done.

    eapply cos_loc_2; eauto; done.

  (* read case *)
  destruct Hfr as (iw & Hrf & Hcopw3 & _).
  destruct Hcoacy; auto; [].
  exists i3.
  eapply star_plus_trans; eauto; [].
  so (eq_ident_dec i3 iw) as [Heq | Hneq].
    subst iw.
    apply star_refl; done.
  apply star_one; [].
  eapply @co_read; eauto; [].
  destruct Hisw3 as (l & Hwrite3).
  #2 eapply wrp_vt; eauto.
    exists ib; auto; done.

    eapply rf_loc_1; eauto; [].
    eapply cos_loc_2; eauto; [].
    apply plus_star; auto; done.
Qed.


(* If  ia -xo->* i -sc->* i' -vo->+ ib
   and H (fence post_x ia)
   and exec ib
   then
       ia -to-> ib
*)
Lemma scs_then_push :
  forall H i i' ia ib,
    trco H
    -> acyclic (co H)
    -> allexec H
    -> isp H ib
    -> xos H ia i
    -> star (sc (ppo H) H) i i'
    -> vop H i' ib
    -> to H ia ib.
Proof.
intros H i1 i2 ia ib Htrco Hcoacy Hallexec Hisp Hxos1 Hscs Hvop2.
assert (star (rsc H) i1 i1) as Hrscs by (apply star_refl).
assert (forall iw ix iy iz, xos H iw ix -> star (rsc H) ix iy -> vop H iy iz -> exec H iw) as Hallexec'.
  intros iw ix iy iz Hwx Hxy Hyz.
  apply allexec_inited; auto; [].
  so (star_plus _#4 Hwx) as [Heq | Hwx'].
    subst ix.
    so (star_plus _#4 Hxy) as [Heq | Hxy'].
      subst iy.
      eapply vop_inited_fst; eauto; done.

      destruct Hxy' as (? & Hrsc & _).
      exact (rsc_inited _#3 Htrco Hrsc andel).

    eapply xop_inited_fst; eauto; done.
revert Hxos1 Hrscs.
generalize i1 at 1 2.
revert ia Hvop2.
induct Hscs.

(* refl *)
intros i2 ia Hvop2 i1 Hxos1 Hrscs.
eapply rscs_then_push; eauto; done.

(* step *)
intros i2 i3 i4 Hsc23 Hscs34 IH ia Hvop2 i1 Hxos1 Hrscs.
destruct Hsc23 as [Hppo23 | Hrsc23].
  destruct Hppo23 as (ic & Hispc & Hvop2c & Hxopc3).
  apply (to_trans _ _ ic); auto; [|].
    eapply rscs_then_push; eauto; done.

    #2 eapply IH; eauto.
      apply plus_star; eauto; done.

      apply star_refl; done.

  eapply IH; eauto; [].
  eapply star_trans; eauto; [].
  apply star_one; assumption.
Qed.


Theorem sc1_acyclic :
  forall H,
    trco H
    -> acyclic (co H)
    -> allexec H
    -> acyclic (sc1 H).
Proof.
intros H Htrco Hcoacy Hallexec.
intros (i & Hscp).
so (sc_cycle_r _ _ _ Htrco Hcoacy Hscp) as (i1 & i2 & Hppo & Hscs).
clear i Hscp.
destruct Hppo as (is & Hisp & Hvop & Hxop).
apply (to_irreflex H is); auto; [].
eapply scs_then_push; eauto; [].
apply plus_star; auto; done.
Qed.


(* Second theorem: sequential consistency for data-race-free programs. *)

Definition sq H i i' :=
  po H i i' \/ sb H i i'.


Lemma sb_vt :
  forall H i i', sb H i i' -> vt H i i'.
Proof.
intros H i i' Hsb.
destruct Hsb as (i'' & Hvop & Hfence & Hpoq).
exists i''.
split; auto; [].
destruct Hpoq as [Heq | Hpo].
  subst; apply star_refl; done.

  apply star_one; eauto with xo; done.
Qed.


(* Main lemma: Show that (sc po) is contained in sq. *)

Lemma racefree_sc2_sq :
  forall H i i',
    trco H
    -> acyclic (co H)
    -> racefree (fun _ => True) H
    -> sc2 H i i'
    -> sq H i i'.
Proof.
intros H i i' Htrco Hcoacy Hracefree Hsc.
destruct Hsc as [Hpo | [Hco | [Hrf | Hfr]]].
  left; assumption.

  so (co_wf_prop _#3 Htrco Hco) as (l & Hwritesto & Hwritesto').
  thus (conflict H i i') as Hconflict by @conflict_ww.
  so (writesto_inited _#3 Htrco Hwritesto) as (p & Hinit).
  so (writesto_inited _#3 Htrco Hwritesto') as (p' & Hinit').
  so (eq_thread_dec p p') as [Heqp | Hneqp].
    subst p'.
    so (program_trichot _#4 Hinit Hinit') as [Hpo | [Heqi | Hpo]].
      left; assumption.

      subst i'.
      destruct Hcoacy.
      exists i.
      apply plus_one; auto; done.

      destruct Hcoacy.
      exists i, i'.
      split; auto; [].
      apply star_one; [].
      apply co_write; [].
      eapply wwp_po; eauto; done.

    so (Hracefree _#4 Hinit Hinit' Hneqp Hconflict I) as [Hsb | Hsb].
      right; assumption.

      destruct Hcoacy.
      exists i, i'.
      split; auto; [].
      apply star_one; [].
      apply co_write; [].
      eapply wwp_vt; eauto using sb_vt; done.

  so (rf_wf_prop _#3 Htrco Hrf) as (l & Hwritesto & Hreads).
  thus (conflict H i i') as Hconflict by @conflict_wr.
  so (writesto_inited _#3 Htrco Hwritesto) as (p & Hinit).
  so (reads_inited _#3 Htrco Hreads) as (p' & Hinit').
  so (eq_thread_dec p p') as [Heqp | Hneqp].
    subst p'.
    so (program_trichot _#4 Hinit Hinit') as [Hpo | [Heqi | Hpo]].
      left; assumption.

      subst i'.
      destruct (rf_irreflex H i); auto; done.

      exfalso.
      eapply rf_po_contra; eauto; done.

    so (Hracefree _#4 Hinit Hinit' Hneqp Hconflict I) as [Hsb | Hsb].
      right; assumption.

      destruct (to_irreflex H i); auto; [].
      #2 apply (to_trans _ _ i').
        apply rf_to; auto; done.

        apply vt_to; auto using sb_vt; [].
        eapply rf_execed_fst; eauto; done.

  destruct Hfr as (iw & Hrf & Hcop & Hneq).
  so (rf_wf_prop _#3 Htrco Hrf) as (l & Hwritestow & Hreads).
  assert (writesto H i' l) as Hwritesto.
    destruct Hwritestow.
    eapply cop_wf_prop_1; eauto; done.
  thus (conflict H i i') as Hconflict by @conflict_rw.
  so (reads_inited _#3 Htrco Hreads) as (p & Hinit).
  so (writesto_inited _#3 Htrco Hwritesto) as (p' & Hinit').
  so (eq_thread_dec p p') as [Heqp | Hneqp].
    subst p'.
    so (program_trichot _#4 Hinit Hinit') as [Hpo | [Heqi | Hpo]].
      left; assumption.

      contradiction.

      exfalso.
      apply Hcoacy; [].
      exists i'.
      eapply plus_trans; eauto; [].
      apply plus_one; [].
      #2 apply (@co_read _#3 i); auto.
        eapply wrp_po; eauto; done.

        intro.
        subst iw.
        destruct Hcoacy; eauto; done.

    so (Hracefree _#4 Hinit Hinit' Hneqp Hconflict I) as [Hsb | Hsb].
      right; assumption.

      exfalso.
      apply Hcoacy; [].
      exists i'.
      eapply plus_trans; eauto; [].
      apply plus_one; [].
      #2 apply (@co_read _#3 i); auto.
        eapply wrp_vt; eauto using sb_vt; done.

        intro.
        subst iw.
        destruct Hcoacy; eauto; done.
Qed.


Lemma sq_acyclic :
  forall H, trco H -> acyclic (sq H).
Proof.
intros H Htrco.
intros (i & Hsqp).
cut (po H i i
     \/
     exists ia ib,
       H {{ ev_fence (fc_post at_x) ib }} 
       /\ poq H i ia
       /\ to H ia ib
       /\ poq H ib i).
  intros [Hpo | (ia & ib & Hfenceb & Hpoa & Hto & Hpob)].
    edestruct po_irreflex; eauto; done.

    destruct (to_irreflex H ia); auto; [].
    eapply to_toq_trans; eauto; [].
    so (poq_trans _#4 Htrco Hpob Hpoa) as Hpoq.
    destruct Hpoq as [Heq | Hpo].
      subst; left; reflexivity.
    right.
    apply xo_to; eauto using to_execed_fst with xo; done.

  revert Hsqp.
  generalize i at 1 3 5.
  intros i' Hsqp.
  so (plus_plusi _#4 Hsqp) as Hsqp'; clear Hsqp.
  induct Hsqp'.
    (* one *)
    intros i1 i2 Hsq.
    destruct Hsq as [Hpo | Hsb].
      left; assumption.
    destruct Hsb as (i & Hvop & Hfence & Hpoq).
    so (vop_to_or_svop _#3 Htrco Hvop) as [Hto | Hsvop].
      right.
      exists i1, i.
      repeat2 split; auto; [].
      left; reflexivity.

      left.
      eapply po_poq_trans; eauto; [].
      eapply sop_po; eauto; done.

    (* step *)
    intros i1 i2 i3 Hsq12 _ IH.
    destruct IH as [Hpo23 | (ia & ib & Hfenceb & Hpoq2a & Htoab & Hpoqb3)].
      destruct Hsq12 as [Hpo12 | Hsb12].
        left; eapply po_trans; eauto; done.
      destruct Hsb12 as (ic & Hvop & Hfencec & Hpoqc2).
      so (vop_to_or_svop _#3 Htrco Hvop) as [Hto1c | Hsvop].
        right.
        exists i1, ic.
        #2 repeat2 split; auto.
          left; reflexivity.

          right.
          eapply poq_po_trans; eauto; done.

        left.
        eapply po_trans; eauto; [].
        eapply po_poq_trans; eauto; [].
        eapply sop_po; eauto; done.

      right.
      destruct Hsq12 as [Hpo12 | Hsb12].
        exists ia, ib.
        repeat2 split; auto; [].
        right.
        eapply po_poq_trans; eauto; done.

        destruct Hsb12 as (ic & Hvop & Hfencec & Hpoqc2).
        exists i1, ib.
        #2 repeat2 split; auto.
          left; reflexivity.

          eapply to_trans; eauto; [].
          assert (toq H ic ia) as Htoca.
            so (poq_trans _#4 Htrco Hpoqc2 Hpoq2a) as [Heq | Hpoca].
              subst; left; reflexivity.
            right.
            apply xo_to; eauto using to_execed_fst with xo; done.
          eapply to_toq_trans; eauto; [].
          apply vop_to; eauto using toq_execed_2, to_execed_fst; done.
Qed.


Theorem sc2_acyclic :
  forall H,
    trco H 
    -> acyclic (co H) 
    -> racefree (fun _ => True) H 
    -> acyclic (sc2 H).
Proof.
intros H Htrco Hcoacy Hracefree.
#2 apply (acyclic_anti _ _ (sq H)).
  intros; apply racefree_sc2_sq; auto; done.

  apply sq_acyclic; auto; done.
Qed.


(* Third theorem: combining the first and second theorems. *)

Definition fpoq F H i i' := i = i' \/ fpo F H i i'.

Definition fsb F H i i' :=
  F i /\ F i' /\ sb H i i'.

Definition exc (F : identifier -> Prop) R i i' :=
  ~ F i
  /\ ~ F i'
  /\ R i i'.

Definition sk F H i i' :=
  sppo H i i' \/ fpo F H i i' \/ fsb F H i i' \/ exc F (co H) i i' \/ exc F (rf H) i i' \/ exc F (fr H) i i'.

Definition rsk F H i i' :=
  fpo F H i i' \/ fsb F H i i' \/ exc F (co H) i i' \/ exc F (rf H) i i' \/ exc F (fr H) i i'.

Definition fsq F H i i' :=
  fpo F H i i' \/ fsb F H i i'.

Definition frsc F H i i' :=
  exc F (co H) i i' \/ exc F (rf H) i i' \/ exc F (fr H) i i'.


Lemma exc_r :
  forall F R i i', exc F R i i' -> R i i'.
Proof.
intros F R i i' (_ & _ & H).
assumption.
Qed.


Lemma fpo_f :
  forall F H i i',
    fpo F H i i'
    -> F i.
Proof.
intros F H i i' Hfpo.
destruct Hfpo; auto.
Qed.


Lemma fpo_po :
  forall F H i i', fpo F H i i' -> po H i i'.
Proof.
intros F H i i' (_ & Hpo).
assumption.
Qed.


Lemma fpoq_poq :
  forall F H i i', fpoq F H i i' -> poq H i i'.
Proof.
intros F H i i' Hfpoq.
destruct Hfpoq; [left | right]; eauto using fpo_po.
Qed.


Lemma fsb_f :
  forall F H i i',
    fsb F H i i'
    -> F i /\ F i'.
Proof.
intros F H i i' Hfsb.
destruct Hfsb as (? & ? & _).
auto.
Qed.


Lemma fsb_sb :
  forall F H i i',
    fsb F H i i'
    -> sb H i i'. 
Proof.
intros F H i i' Hfsb.
destruct Hfsb as (_ & _ & ?).
auto.
Qed.


Lemma rsk_fsq_or_frsc :
  forall F H i i',
    rsk F H i i'
    -> fsq F H i i' \/ frsc F H i i'.
Proof.
intros F H i i' Hrsk.
destruct Hrsk as [Hfpo | [Hfsb | Hrrqc]].
  left; left; assumption.

  left; right; assumption.

  right; assumption.
Qed.


Lemma fsq_sq :
  forall F H i i',
    fsq F H i i'
    -> sq H i i'.
Proof.
intros F H i i' Hfsq.
destruct Hfsq.
  left; eauto using fpo_po; done.

  right; eauto using fsb_sb; done.
Qed.


Lemma fsq_f :
  forall F H i i',
    fsq F H i i'
    -> F i.
Proof.
intros F H i i' Hfsq.
destruct Hfsq.
  eapply fpo_f; eauto; done.

  refine (fsb_f _#5 andel); eauto; done.
Qed.


Lemma frsc_nf :
  forall F H i i',
    frsc F H i i'
    -> ~ F i /\ ~ F i'.
Proof.
intros F H i i' Hfrsc.
destruct Hfrsc as [HH | [HH | HH]]; destruct HH as (? & ? & _); auto; done.
Qed.


Lemma frsc_rsc :
  forall F H i i', frsc F H i i' -> rsc H i i'.
Proof.
intros F H i i' [Hco | [Hrf | Hfr]].
  left; eapply exc_r; eauto; done.

  right; left; eapply exc_r; eauto; done.

  right; right; eapply exc_r; eauto; done.
Qed.


Lemma frscs_rscs :
  forall F H i i', star (frsc F H) i i' -> star (rsc H) i i'.
Proof.
intros F H i i' Hfrscs.
refine (star_mono _#6 Hfrscs); [].
intros; eapply frsc_rsc; eauto; done.
Qed.


Lemma frscp_nf :
  forall F H i i',
    plus (frsc F H) i i'
    -> ~ F i /\ ~ F i'.
Proof.
intros F H i i' Hfrscp.
split.
  destruct Hfrscp as (? & HH & _).
  exact (frsc_nf _#4 HH andel).

  so (plus_plusr _#4 Hfrscp) as (? & _ & HH).
  exact (frsc_nf _#4 HH ander).
Qed.


Lemma frscp_rscp :
  forall F H i i', plus (frsc F H) i i' -> plus (rsc H) i i'.
Proof.
intros F H i i' Hfrscs.
refine (plus_mono _#6 Hfrscs); [].
intros; eapply frsc_rsc; eauto; done.
Qed.


Lemma rsk_sk :
  forall F H i i',
    rsk F H i i' -> sk F H i i'.
Proof.
intros F H i i' Hrsk.
right; assumption.
Qed.


Lemma rsks_sks :
  forall F H i i',
    star (rsk F H) i i'
    -> star (sk F H) i i'.
Proof.
intros F H i i' Hrsks.
apply (star_mono _ (rsk F H)); auto using rsk_sk.
Qed.


(* First we establish that an sk cycle contains a ppo edge. *)

(* If there is a cycle in rsk, then there is a cycle in either rsc or sq.  Neither can happen, as shown in parts 1 and 2. *)
Lemma rsk_acyclic :
  forall F H,
    trco H
    -> acyclic (co H)
    -> acyclic (rsk F H).
Proof.
intros F H Htrco Hcoacy.
intros (i & Hrsqp).
cut (plus (frsc F H) i i
     \/ exists i', F i /\ plus (sq H) i i' /\ star (frsc F H) i' i).
  intros [Hfrscp | (i' & Hf & Hsqp & Hfrscs)].
    destruct (rsc_acyclic _ Htrco Hcoacy); [].
    exists i.
    eapply frscp_rscp; eauto; done.

    so (star_plus _#4 Hfrscs) as [Heq | Hfrscp].
      subst i'.
      destruct (sq_acyclic _ Htrco); eauto; done.

      so (frscp_nf _#4 Hfrscp) as (_ & Hnf).
      contradiction.
revert Hrsqp.
generalize i at 1 3 5 6.
intros i' Hrskp.
so (plus_plusi _#4 Hrskp) as Hrskp'; clear Hrskp.
induct Hrskp'.
(* one *)
intros i1 i2 Hrsk.
destruct Hrsk as [Hpo | [Hsb | Hrsc]].
  right.
  exists i2.
  destruct Hpo as (Hf & Hpo).
  #2 repeat2 split; auto.
    apply plus_one; [].
    left; assumption.

    apply star_refl; done.

  right.
  exists i2.
  destruct Hsb as (Hf & _ & Hsb).
  #2 repeat2 split; auto.
    apply plus_one; [].
    right; assumption.

    apply star_refl; done.

  change (frsc F H i1 i2) in Hrsc.
  left.
  apply plus_one; assumption.

(* step *)
intros i1 i2 i3 H12 H23 IH.
assert ((F i1 /\ sq H i1 i2) \/ frsc F H i1 i2) as [(Hf & Hsq) | Hfrsc].
  destruct H12 as [Hfpo12 | [Hsb12 | Hrsc12]].
    left.
    split.
      eapply fpo_f; eauto; done.

      left.
      eapply fpo_po; eauto; done.

    left.
    split.
      refine (fsb_f _#5 andel); eauto; done.
      
      right.
      eapply fsb_sb; eauto; done.

    right; assumption.

  destruct IH as [Hfrscp | (i4 & _ & Hsqp & Hfrscs)].
    right.
    exists i2.
    #2 repeat2 split; auto.
      apply plus_one; auto; done.

      apply plus_star; auto; done.

    right.
    exists i4.
    repeat2 split; auto; [].

    eapply star_plus_trans; eauto; [].
    apply star_one; auto; done.

  destruct IH as [Hfrscp | (i4 & HH & Hsqp & Hfrscs)].
    left.
    eapply plus_trans; eauto; [].
    apply plus_one; auto; done.

    exfalso.
    so (frsc_nf _#4 Hfrsc ander) as Hnf.
    contradiction.
Qed.


Lemma sk_cycle_sppo :
  forall F H i,
    trco H
    -> acyclic (co H)
    -> plus (sk F H) i i
    -> exists i1 i2,
         sppo H i1 i2
         /\ star (sk F H) i2 i1.
Proof.
intros F H i Htrco Hcoacy Hcycle.
destruct Hcycle as (i' & Hsk & Hsks).
destruct Hsk as [Hppo | Hrsk].
  exists i, i'; splitall; auto; done.

  change (rsk F H i i') in Hrsk.
  thus (plus (rsk F H) i i') as Hacc by plus_one; clear Hrsk.
  remember i as i0 eqn:Heqi0 in Hacc.
  revert i0 Heqi0 Hacc.
  induct Hsks.
  (* refl *)
  intros i i0 Heqi0 Hacc.
  subst i0.
  destruct (rsk_acyclic F _ Htrco Hcoacy); eauto; done.

  (* step *)
  intros i1 i2 i3 H12 H23 IH i0 Heqi0 Hacc.
  subst i0.
  destruct H12 as [Hppo | Hrsk].
    exists i1, i2.
    split; auto; [].
    eapply star_trans; eauto; [].
    #2 apply (star_mono _ (rsk F H)).
      intros; right; auto; done.

      apply plus_star; auto; done.
      
    apply (IH _ (eq_refl _)); [].
    eapply plus_trans; eauto using plus_one; done.
Qed.         


Lemma partition_rsks :
  forall F H i i',
    trco H
    -> star (rsk F H) i i'
    -> exists i'',
         star (fsq F H) i i'' /\ star (frsc F H) i'' i'.
Proof.
intros F H i i' Htrco Hrsks.
induct Hrsks.

(* refl *)
intros i.
exists i; auto using star_refl; done.

(* step *)
intros i1 i2 i3 H12 _ IH.
destruct IH as (i4 & H24 & H43).
so (rsk_fsq_or_frsc _#4 H12) as [Hfsq | Hfrsc].
  exists i4.
  split; auto; [].
  eapply star_step; eauto; done.

  so (star_plus _#4 H24) as [Heq | Hfsqp].
    subst i4.
    exists i1.
    split; [apply star_refl; done |].
    eapply star_step; eauto; done.

    destruct Hfsqp as (? & Hfsq & _).
    so (fsq_f _#4 Hfsq).
    so (frsc_nf _#4 Hfrsc ander).
    contradiction.
Qed.


Lemma rsks_then_push :
  forall F H i i' ia ib,
    trco H
    -> acyclic (co H)
    -> allexec H
    -> H {{ ev_fence (fc_post at_x) ia }}
    -> isp H ib
    -> po H ia i
    -> star (rsk F H) i i'
    -> vop H i' ib
    -> to H ia ib.
Proof.
intros F H i1 i3 ia ib Htrco Hcoacy Hallexec Hfencea Hisp Hpo1 Hrsks Hvop.
thus (exec H ib) as Hexecb by allexec_is.
so (partition_rsks _#4 Htrco Hrsks) as (i2 & Hsq & Hrscs).
assert (exists ic, toq H ia ic /\ H {{ ev_fence (fc_post at_x) ic }} /\ poq H ic i2) as (ic & Htoqac & Hfencec & Hpoc2).
  assert (poq H i1 i2
          \/ exists i4 i5, poq H i1 i4 /\ toq H i4 i5 /\ H {{ ev_fence (fc_post at_x) i5 }} /\ poq H i5 i2) as [Hpoq | (i4 & i5 & H14 & H45 & Hfence5 & H52)].
    clear i3 ia ib Hfencea Hisp Hpo1 Hrsks Hvop Hexecb Hrscs.
    induct Hsq.
      (* refl *)
      intro i.
      left; left; reflexivity.

      (* step *)
      intros i1 i2 i3 H12 _ IH.
      #4 destruct H12 as [Hpo12 | Hsb]; destruct IH as [Hpoq23 | (i4 & i5 & Hpoq24 & Htoq45 & Hfence5 & Hpoq53)].
        left; right.
        eapply po_poq_trans; eauto using fpo_po; done.

        right.
        exists i4, i5.
        repeat2 split; auto; [].
        right.
        eapply po_poq_trans; eauto using fpo_po; done.

        so (fsb_sb _#4 Hsb) as (i6 & Hvop & Hfence6 & Hpoq62).
        right.
        exists i1, i6.
        #3 repeat2 split; auto.
          left; reflexivity.

          right.
          apply vop_to'; eauto using allexec_inited, vop_inited_fst; done.

          eapply poq_trans; eauto; done.

        right.
        so (fsb_sb _#4 Hsb) as (i6 & Hvop & Hfence6 & Hpoq62).
        exists i1, i5.
        #2 repeat2 split; auto.
          left; reflexivity.

          right.
          eapply to_toq_trans; eauto; [].
          #2 apply (to_toq_trans _ _ i6); auto.
            apply vop_to'; eauto using allexec_inited, vop_inited_fst; done.

            thus (poq H i6 i4) as Hpoq64 by poq_trans.
            destruct Hpoq64 as [Heq | Hpo64].
              subst; left; reflexivity.
            
              right.
              apply xo_to; eauto using allexec_po_snd with xo; done.

    exists ia.
    #2 repeat2 split; auto.
      left; reflexivity.

      right.
      eapply po_poq_trans; eauto; done.

    exists i5.
    repeat2 split; auto; [].
    right.
    eapply to_toq_trans; eauto; [].
    apply xo_to'; eauto using po_poq_trans, allexec_po_fst with xo; done.
eapply toq_to_trans; eauto; [].
#2 apply (rscs_then_push _ i2 i3); eauto using frscs_rscs.
  apply allexec_inited; auto; [].
  eapply fence_wf_prop; eauto; done.

  destruct Hpoc2 as [Heq | Hpoc2].
    subst; apply star_refl; done.

    apply star_one; eauto with xo; done.
Qed.


Lemma sks_then_push :
  forall F H i i' ia ib,
    trco H
    -> acyclic (co H)
    -> allexec H
    -> H {{ ev_fence (fc_post at_x) ia }}
    -> isp H ib
    -> po H ia i
    -> star (sk F H) i i'
    -> vop H i' ib
    -> to H ia ib.
Proof.
intros F H i1 i2 ia ib Htrco Hcoacy Hallexec Hfence Hisp Hpo1 Hsks Hvop2.
assert (star (rsk F H) i1 i1) as Hrsks by (apply star_refl).
revert Hpo1 Hrsks.
generalize i1 at 1 2.
revert ia Hfence Hvop2.
induct Hsks.

(* refl *)
intros i2 ia Hfence Hvop2 i1 Hpo1 Hrsks.
eapply rsks_then_push; eauto; done.

(* step *)
intros i2 i3 i4 Hsc23 Hscs34 IH ia Hfence Hvop2 i1 Hpo1 Hrscs.
destruct Hsc23 as [Hppo23 | Hrsc23].
  destruct Hppo23 as (ic & Hispc & Hfencec & Hvop2c & Hxopc3).
  apply (to_trans _ _ ic); auto; [|].
    eapply rsks_then_push; eauto; done.

    eapply IH; eauto using xop_po, star_refl; done.

  eapply IH; eauto; [].
  eapply star_trans; eauto; [].
  apply star_one; assumption.
Qed.


Lemma sk_acyclic :
  forall F H,
    trco H
    -> acyclic (co H)
    -> allexec H
    -> acyclic (sk F H).
Proof.
intros F H Htrco Hcoacy Hallexec.
intros (i & Hskp).
so (sk_cycle_sppo _ _ _ Htrco Hcoacy Hskp) as (i1 & i2 & Hppo & Hsks).
clear i Hskp.
destruct Hppo as (is & Hisp & Hfence & Hvop & Hxop).
apply (to_irreflex H is); auto; [].
eapply sks_then_push; eauto; [].
apply xop_po; auto; done.
Qed.


Lemma racefree_sc3_sk :
  forall F H i i',
    (forall i, decidable (F i))
    -> trco H
    -> acyclic (co H)
    -> racefree F H
    -> (forall i i' l, accesses H i l -> accesses H i' l -> F i -> F i')
    -> sc3 F H i i'
    -> sk F H i i'.
Proof.
intros F H i i' Hdec Htrco Hcoacy Hracefree Hdisjoint Hsc.
destruct Hsc as [[Hfpo | Hppo] | [Hco | [Hrf | Hfr]]].
  left; assumption.

  right; left; assumption.

  so (co_wf_prop _#3 Htrco Hco) as (l & Hwritesto & Hwritesto').
  so (Hdec i) as [Hf | Hnf].
    thus (conflict H i i') as Hconflict by @conflict_ww.
    so (writesto_inited _#3 Htrco Hwritesto) as (p & Hinit).
    so (writesto_inited _#3 Htrco Hwritesto') as (p' & Hinit').
    so (eq_thread_dec p p') as [Heqp | Hneqp].
      subst p'.
      so (program_trichot _#4 Hinit Hinit') as [Hpo | [Heqi | Hpo]].
        right; left.
        split; auto; done.
  
        subst i'.
        destruct Hcoacy.
        exists i.
        apply plus_one; auto; done.
  
        destruct Hcoacy.
        exists i, i'.
        split; auto; [].
        apply star_one; [].
        apply co_write; [].
        eapply wwp_po; eauto; done.
  
      so (Hracefree _#4 Hinit Hinit' Hneqp Hconflict Hf) as [Hsb | Hsb].
        right; right; left.
        repeat2 split; eauto using Hdisjoint, writesto_accesses; done.
  
        destruct Hcoacy.
        exists i, i'.
        split; auto; [].
        apply star_one; [].
        apply co_write; [].
        eapply wwp_vt; eauto using sb_vt; done.

    assert (~ F i') as Hnf' by (intro; destruct Hnf; apply (Hdisjoint i' _ l); eauto using writesto_accesses).
    right; right; right; left.
    unfold exc; auto; done.

  so (rf_wf_prop _#3 Htrco Hrf) as (l & Hwritesto & Hreads).
  so (Hdec i) as [Hf | Hnf].
    thus (conflict H i i') as Hconflict by @conflict_wr.
    so (writesto_inited _#3 Htrco Hwritesto) as (p & Hinit).
    so (reads_inited _#3 Htrco Hreads) as (p' & Hinit').
    so (eq_thread_dec p p') as [Heqp | Hneqp].
      subst p'.
      so (program_trichot _#4 Hinit Hinit') as [Hpo | [Heqi | Hpo]].
        right; left.
        split; auto; done.
  
        subst i'.
        destruct (rf_irreflex H i); auto; done.
  
        exfalso.
        eapply rf_po_contra; eauto; done.
  
      so (Hracefree _#4 Hinit Hinit' Hneqp Hconflict Hf) as [Hsb | Hsb].
        right; right; left.
        repeat2 split; eauto using Hdisjoint, writesto_accesses, reads_accesses; done.
  
        destruct (to_irreflex H i); auto; [].
        #2 apply (to_trans _ _ i').
          apply rf_to; auto; done.
  
          apply vt_to; auto using sb_vt; [].
          eapply rf_execed_fst; eauto; done.
  
    assert (~ F i') as Hnf' by (intro; destruct Hnf; apply (Hdisjoint i' _ l); eauto using reads_accesses, writesto_accesses).
    right; right; right; right; left.
    unfold exc; auto; done.

  so Hfr as (iw & Hrf & Hcop & Hneq).
  so (Hdec i) as [Hf | Hnf].
    so (rf_wf_prop _#3 Htrco Hrf) as (l & Hwritestow & Hreads).
    assert (writesto H i' l) as Hwritesto.
      destruct Hwritestow.
      eapply cop_wf_prop_1; eauto; done.
    thus (conflict H i i') as Hconflict by @conflict_rw.
    so (reads_inited _#3 Htrco Hreads) as (p & Hinit).
    so (writesto_inited _#3 Htrco Hwritesto) as (p' & Hinit').
    so (eq_thread_dec p p') as [Heqp | Hneqp].
      subst p'.
      so (program_trichot _#4 Hinit Hinit') as [Hpo | [Heqi | Hpo]].
        right; left.
        split; auto; done.
  
        contradiction.
  
        exfalso.
        apply Hcoacy; [].
        exists i'.
        eapply plus_trans; eauto; [].
        apply plus_one; [].
        #2 apply (@co_read _#3 i); auto.
          eapply wrp_po; eauto; done.
  
          intro.
          subst iw.
          destruct Hcoacy; eauto; done.
  
      so (Hracefree _#4 Hinit Hinit' Hneqp Hconflict Hf) as [Hsb | Hsb].
        right; right; left.
        repeat2 split; eauto using Hdisjoint, writesto_accesses, reads_accesses; done.
  
        exfalso.
        apply Hcoacy; [].
        exists i'.
        eapply plus_trans; eauto; [].
        apply plus_one; [].
        #2 apply (@co_read _#3 i); auto.
          eapply wrp_vt; eauto using sb_vt; done.
  
          intro.
          subst iw.
          destruct Hcoacy; eauto; done.

    so (cop_wf_prop _#3 Htrco Hcop) as (l & Hwritesto & Hwritesto').
    so (rf_loc_1 _#4 Htrco Hrf Hwritesto) as Hreads.
    assert (~ F i') as Hnf' by (intro; destruct Hnf; apply (Hdisjoint i' _ l); eauto using reads_accesses, writesto_accesses).
    right; right; right; right; right.
    unfold exc; auto; done.
Qed.


Theorem sc3_acyclic :
  forall F H,
    (forall i, decidable (F i))
    -> trco H
    -> acyclic (co H)
    -> allexec H
    -> racefree F H
    -> (forall i i' l, accesses H i l -> accesses H i' l -> F i -> F i')
    -> acyclic (sc3 F H).
Proof.
intros F H Hdec Htrco Hcoacy Hallexec Hracefree Hdisjoint.
#2 apply (acyclic_anti _ _ (sk F H)).
  intros; apply racefree_sc3_sk; auto; done.

  apply sk_acyclic; auto; done.
Qed.


(* Fourth theorem: existence of a sequentially consistent order *)

Definition sequentializable R H :=
  exists Q,
    irreflexive _ Q
    /\
    transitive _ Q
    /\
    (* Q is total *)
    (forall i i', inited H i -> inited H i' -> Q i i' \/ i = i' \/ Q i' i)
    /\
    (* Q respects R *)
    (forall i i', R i i' -> Q i i')
    /\
    (* Each read is satisfied by the most recent write to its location. *)
    (forall iw ir,
       rf H iw ir
       -> Q iw ir
          /\ ~ exists l iw', reads H ir l /\ writesto H iw' l /\ Q iw iw' /\ Q iw' ir).  


(* Messy, but not hard or interesting. *)
Lemma topological_sort :
  forall (T : Set) (R : T -> T -> Prop) (S : T -> Prop),
    (forall (x y : T), decidable (x = y))
    -> (forall x y, decidable (R x y))
    -> irreflexive _ R
    -> transitive _ R
    -> itemizable S
    -> exists Q,
         (forall x y, decidable (Q x y))
         /\ irreflexive _ Q
         /\ transitive _ Q
         /\ (forall x y, S x -> S y -> Q x y \/ x = y \/ Q y x)
         /\ (forall x y, R x y -> Q x y).
Proof.
intros T R S HdecT HdecR Hirrefl Htrans Hitem.
destruct Hitem as (L & Hiniff).
cut (exists Q,
       (forall x y, decidable (Q x y))
       /\ irreflexive _ Q
       /\ transitive _ Q
       /\ (forall x y, In x L -> In y L -> Q x y \/ x = y \/ Q y x)
       /\ (forall x y, R x y -> Q x y)).
  intros (Q & Hcond).
  destruct Hcond as (Hdec' & Hirrefl' & Htrans' & Htotal & Hresp).
  exists Q; repeat2 split; eauto; [].
  intros.
  apply Htotal; apply Hiniff; auto; done.
clear S Hiniff.
generalize L at 2.
induction L as [| x L IH1].

(* nil *)
intro L.
exists R.
do2 4 split; auto; [].
intros x y Hin _.
destruct Hin; done.

(* cons *)
clear HdecR Hirrefl Htrans.
intro M.
destruct (IH1 M) as (Q & Hdec & Hirrefl & Htrans & Htotal & Hresp).
clear IH1.
cut (exists Q',
       (forall x y, decidable (Q' x y))
       /\ irreflexive _ Q'
       /\ transitive _ Q'
       /\ (forall y, In y M -> Q' x y \/ x = y \/ Q' y x)
       /\ (forall x y, Q x y -> Q' x y)).
  intros (Q' & Hdec' & Hirrefl' & Htrans' & Htotal' & Hresp').
  exists Q'.
  do2 4 split; auto; [].
  intros x' y HinL HinM.
  destruct HinL as [Heq | HinL].
    subst x'.
    apply Htotal'; auto; done.

    so (Htotal _ _ HinL HinM) as [|[|]]; auto using Hresp'.
induction M as [| y M IH2].
  (* nil *)
  exists Q.
  do2 4 split; auto; [].
  intros y Hin.
  destruct Hin; done.

  (* cons *)
  lapply IH2.
    clear Hdec Hirrefl Htrans Htotal Hresp IH2.
    intros (Q' & Hdec & Hirrefl & Htrans & Htotal & Hresp).
    so (dec_or _ _ (Hdec x y) (dec_or _ _ (HdecT x y) (Hdec y x))) as [Hrel | Hnrel].
      exists Q'.
      do2 4 split; auto; [].
      intros y' HinM.
      destruct HinM as [Heq | HinM].
        subst y'.
        assumption.

        apply Htotal; auto; done.

      exists (fun x' y' => ((x' = x \/ Q' x' x) /\ (Q' y y' \/ y = y')) \/ Q' x' y').
      #5 do2 4 split.
        auto using dec_or, dec_and; done.

        #2 intros (z & [([Hzx | Hxz] & [Hyz | Hyz]) | Hzz]); try (subst; destruct Hnrel; auto; done).
          destruct Hnrel.
          right; right.
          eapply Htrans; eauto; done.

          destruct Hirrefl; eauto; done.

        intros u v w Huv Hvw.
        #4 destruct Huv as [([Hux | Hux] & [Hyv | Hyv]) | Huv]; destruct Hvw as [([Hvx | Hvx] & [Hyw | Hyw]) | Hvw]; subst; eauto using Htrans.
          left; split; left; eauto using Htrans; done.

          left; split; [right | left]; eauto using Htrans; done.

          left; split; [right | left]; eauto using Htrans; done.

          left; split; right; eauto using Htrans; done.
          
        intros y' HinM.
        destruct HinM as [Heq | HinM].
          subst y'.
          left; left; split; [left | right]; reflexivity.

          so (Htotal _ HinM) as [Hq | [Heq | Hq]].
          left; right; assumption.

          right; left; assumption.

          right; right; right; assumption.

        intros x' y' Hq.
        right.
        apply Hresp; auto; done.

    intros x' y' HinL HinM.
    apply Htotal; auto; [].
    right; assumption.
Qed.


Theorem sequential_consistency :
  forall R H,
    (forall i i', decidable (R i i'))
    -> (exists U, forall i i', R i i' -> In i U /\ In i' U)
    -> trco H
    -> allexec H
    -> (forall i i' l, writesto H i l -> writesto H i' l -> cop H i i' \/ i = i' \/ cop H i' i)
    -> acyclic (sc R H)
    -> sequentializable R H.
Proof.
intros R H Hdec (U & Huniv) Htrco Hallexec Hcototal Hacy.
assert (exists Q,
          irreflexive _ Q
          /\ transitive _ Q
          /\ (forall i i', inited H i -> inited H i' -> Q i i' \/ i = i' \/ Q i' i)
          /\ (forall i i', sc R H i i' -> Q i i'))
  as (Q & Hirrefl & Htrans & Htotal & Hresp).
  #6 exploit (topological_sort _ (plus (sc R H)) (inited H)) as Hcond.
    exact (eq_dec_decidable _ eq_ident_dec).

    intros i1 i2.
    apply (itemizable_decidable _ (fun i2 => plus (sc R H) i1 i2)); [| exact (eq_dec_decidable _ eq_ident_dec)].
    #3 apply (itemizable_plus_right_finite _ _ (U ++ ids H)).
      exact (eq_dec_decidable _ eq_ident_dec).

      intros x y Hsc.
      apply in_or_app; [].
      destruct Hsc as [Hr | Hrsc].
        left.
        exact (Huniv _ _ Hr andel).

        right.
        apply init_ids; [].
        destruct Hrsc as [Hco | [Hrf | Hfr]].
          exact (co_inited _#3 Htrco Hco andel).

          exact (rf_inited _#3 Htrco Hrf andel).

          exact (fr_inited _#3 Htrco Hfr andel).

        intro i3.
        #4 do2 3 (apply itemizable_or).
          #2 apply (itemizable_from_univ _ _ U).
            intros; apply Hdec; done.

            intros; eapply Huniv; eauto; done.

          apply co_right_itemizable; auto; done.

          apply rf_itemizable_right; done.

          apply fr_itemizable_right; auto; done.

    exact Hacy.

    unfold transitive.
    intros.
    eapply plus_trans; eauto; done.

    apply inited_itemizable; done.

    destruct Hcond as (Q & _ & Hirrefl & Htrans & Htotal & Hresp).
    exists Q.
    do2 3 split; auto; [].
    intros; apply Hresp; auto; [].
    apply plus_one; auto; done.

  exists Q.
  #2 do2 4 split; auto.
    intros i i' Hr.
    apply Hresp; [].
    left; assumption.

    intros iw ir Hrf.
    split.
      apply Hresp; [].
      right; right; left; assumption.

      intros (l & iw' & Hread & Hwritesto' & Hqw & Hqr).
      apply Hirrefl; [].
      exists ir.
      eapply Htrans; eauto; [].
      apply Hresp; [].
      right; right; right.

      exists iw.
      do2 2 split.
        assumption.

        so (rf_loc_2 _#4 Htrco Hrf Hread) as Hwritesto.
        so (Hcototal _#3 Hwritesto Hwritesto') as [Hcop | [Heq | Hcop]].
          assumption.

          exfalso.
          subst iw'.
          apply Hirrefl; eauto; done.

          exfalso.
          apply Hirrefl; [].
          exists iw.
          eapply Htrans; eauto; [].
          so (plus_plusi _#4 Hcop) as Hcop'.
          #2 induct Hcop'.
            intros.
            apply Hresp; [].
            right; left; assumption.

            intros; eapply Htrans; eauto; [].
            apply Hresp; [].
            right; left; assumption.

        intro Heq.
        subst iw'.
        apply Hirrefl; eauto; done.
Qed.


Lemma rwf_descendant :
  forall H i,
    trco H
    -> isw H i
    -> exists i',
         rwfs H i i'
         /\ ~ exists i'', rwf H i' i''.
Proof.
intros H i Htrco.
wfinduct i using (well_founded_tot _ Htrco).
unfold transpose.
intros i IH Hisw.
so (itemizable_decidable_exists _ _ (rwf_itemizable_right H i)) as [(i' & Hrwf) | Hnrwf].
  #3 exploit (IH i') as (i'' & Hrwfs & Hlast).
    apply rf_to; auto; [].
    apply rwf_rf; auto; done.

    destruct Hrwf as (_ & (l & v & His)).
    unfold isw; eauto using writes_rw; done.

    exists i''.
    split; eauto; [].
    eapply star_step; eauto; done.

  exists i.
  split; auto; [].
  apply star_refl; done.
Qed.


(* This is extremely messy.  It's similar to topological_sort, but enough different to be best done again.
   The key difficulty is that we can't just add coherence edges between unrelated writes; we need to add them
   in such a way as to respect rwf chains.
*)
Lemma complete_co_l :
  forall H l,
    trco H
    -> acyclic (co H)
    -> exists H' H'',
         H' = H;; H''
         /\ trco H'
         /\ acyclic (co H')
         /\ (forall i i', writesto H' i l -> writesto H' i' l -> cop H' i i' \/ i = i' \/ cop H' i' i)
         /\ Forall (fun h => exists i i', h = ev_co i i') H''.
Proof.
intros H l Htrco Hcoacy.
so (writesto_itemizable _ l Htrco) as Hitem.
destruct Hitem as (L & Hiniff).
cut (exists H',
       trco (H;; H')
       /\ acyclic (co (H;; H'))
       /\ (forall i i', In i L -> In i' L -> cop (H;; H') i i' \/ i = i' \/ cop (H;; H') i' i)
       /\ Forall (fun h => exists i i', h = ev_co i i') H').
  intros (H' & Htrco' & Hcoacy' & Htotal & Hallco).
  exists (H;; H'), H'.
  repeat2 split; auto; [].
  intros i i' Hwrite Hwrite'.
  assert (~ H' {{ ev_is i i' | i i' }}).
    intros (i1 & i2 & His).
    so (Forall_forall _#3 andel Hallco _ His) as Heq.
    simpl in Heq; destruct_all Heq; discriminate Heq.
  apply Htotal; apply Hiniff; eapply writesto_truncate; eauto; done.
so (In_iff_impl_Forall _#3 Hiniff) as Hallwrite.
clear Hiniff.
so Hallwrite as Hallwrite'.
revert Hallwrite'.
generalize L at 1 3.
induct Hallwrite.

(* nil *)
intros L _.
exists nil.
do2 3 split; autorewrite with canonlist; auto; [].
intros ? ? Hin.
destruct Hin; done.

(* cons *)
intros i1 L Hwrite1 _ IH1 M HallwriteM.
destruct (IH1 _ HallwriteM) as (H1 & Htrco1 & Hcoacy1 & Htotal1 & Hallco1).
clear IH1.
cut (exists H2,
       trco (H;; H1;; H2)
       /\ acyclic (co (H;; H1;; H2))
       /\ (forall i2, In i2 M -> cop (H;; H1;; H2) i1 i2 \/ i1 = i2 \/ cop (H;; H1;; H2) i2 i1)
       /\ Forall (fun h => exists i i', h = ev_co i i') H2).
  intros (H2 & Htrco2 & Hcoacy2 & Htotal2 & Hallco2).
  exists (H1;; H2).
  autorewrite with canonlist.
  #2 repeat2 split; auto.
    intros i1' i2 [Heq | HinL] HinM.
      subst i1'.
      apply Htotal2; auto; done.

      so (Htotal1 _ _ HinL HinM) as [Hcop | [Heq | Hcop]]; auto using cop_app_incl; done.

    apply Forall_app; auto; done.
clear Htotal1.
induct HallwriteM.
  (* nil *)
  exists nil.
  do2 3 split; autorewrite with canonlist; auto; [].
  intros ? Hin.
  destruct Hin; done.

  (* cons *)
  intros i2 M Hwrite2 _ IH2.
  destruct IH2 as (H2 & Htrco2 & Hcoacy2 & Htotal2 & Hallco2).
  assert (~ H1 {{ ev_is i a | i a }}) as Hnis1.
    intros (? & ? & Hin).
    so (Forall_forall _#3 andel Hallco1 _ Hin) as Heq.
    simpl in Heq; destruct_all Heq; discriminate Heq.
  assert (~ H2 {{ ev_is i a | i a }}) as Hnis2.
    intros (? & ? & Hin).
    so (Forall_forall _#3 andel Hallco2 _ Hin) as Heq.
    simpl in Heq; destruct_all Heq; discriminate Heq.
  assert (~ H1 {{ ev_rf i i' | i i' }}) as Hnrf1.
    intros (? & ? & Hin).
    so (Forall_forall _#3 andel Hallco1 _ Hin) as Heq.
    simpl in Heq; destruct_all Heq; discriminate Heq.
  assert (~ H2 {{ ev_rf i i' | i i' }}) as Hnrf2.
    intros (? & ? & Hin).
    so (Forall_forall _#3 andel Hallco2 _ Hin) as Heq.
    simpl in Heq; destruct_all Heq; discriminate Heq.
  match goal with |- ?G => assert (cop (H;; H1;; H2) i1 i2 \/ i1 = i2 \/ cop (H;; H1;; H2) i2 i1 -> G) as Hrelprop end.
    exists H2.
    repeat2 split; auto; [].
    intros i2' Hin.
    destruct Hin as [Heq | Hin].
      subst; assumption.

      so (Htotal2 _ Hin) as [Hcop | [Heq | Hcop]]; auto using cop_cons_incl; done.
  assert (forall i3 i3d i4,
            trco (H;; H1;; H2; ev_co i3 i4)
            -> rwfs H i3 i3d
            -> ~ (exists i, rwf H i3d i)
            -> ~ cos (H;; H1;; H2) i4 i3d
            -> acyclic (co (H;; H1;; H2; ev_co i3 i4))) as Hprop.
    intros i3 i3d i4 Htrco2' Hrwfsd Hlast Hncos.
    intros (i & Hcop).
    #4 exploit (excise_path _ (co (H;; H1;; H2; ev_co i3 i4)) (co (H;; H1;; H2)) (rwfs H i3) i4 i i) as [Hcop' | (i' & Hcos3 & Hrwfs & Hcos4)].
      intros x y Hxy.
      so (co_cons_invert _#5 Htrco2' Hxy) as [Hco | (Hrwfs & Heq)]; auto; [].
      right; split; auto; [].
      eauto using rwfs_truncate; done.

      assumption.

      destruct Hcoacy2; eauto; done.

      destruct Hncos.
      eapply star_trans; eauto; [].
      eapply star_trans; eauto; [].
      apply rwfs_cos; auto; [].
      eauto using rwfs_app_incl, rwfs_last; done.
  so (rwf_descendant _ _ Htrco (eintro _ Hwrite1)) as (i1d & Hrwfsd1 & Hlast1).
  so (cos_decidable _ i2 i1d Htrco2) as [Hcos21d | Hncos21d].
    so (rwf_descendant _ _ Htrco (eintro _ Hwrite2)) as (i2d & Hrwfsd2 & Hlast2).
    so (eq_ident_dec i1d i2d) as [Heqd | Hneqd].
      subst i2d.
      apply Hrelprop; [].
      so (rwf_chain_trichot _#4 Htrco Hrwfsd1 Hrwfsd2) as [Hrwfp | [Heq | Hrwfp]].
        left.
        repeat (apply cop_app_incl); [].
        apply rwfp_cop; auto; done.

        right; left; assumption.

        right; right.
        repeat (apply cop_app_incl); [].
        apply rwfp_cop; auto; done.

      so (cos_decidable _ i1 i2d Htrco2) as [Hcos12d | Hncos12d].
        so (star_plus _#4 Hcos12d) as [Heq | Hcop12d].
          subst i2d.
          apply Hrelprop; [].
          so (star_plus _#4 Hrwfsd2) as [Heq | Hrwfpd2].
            subst i2.
            right; left; reflexivity.
  
            right; right.
            repeat (apply cop_app_incl); [].
            apply rwfp_cop; assumption.
  
          so (rwfs_cop_group _#4 Htrco2 (rwfs_app_incl _#4 (rwfs_app_incl _#4 Hrwfsd1)) Hcop12d) as [(_ & Hrwfsd) | Hcopd1].
            so (star_plus _#4 Hrwfsd) as [Heq | Hrwfdd].
              subst; destruct Hneqd; auto; done.

              destruct Hlast2.
              destruct Hrwfdd as (i3d & Hrwf3 & _).
              exists i3d.
              eauto using rwf_truncate; done.

            so (star_plus _#4 Hcos21d) as [Heq | Hcop21d].
              subst i1d.
              apply Hrelprop; [].
              so (star_plus _#4 Hrwfsd1) as [Heq | Hrwfpd1].
                subst i1.
                right; left; reflexivity.

                left.
                repeat (apply cop_app_incl); [].
                apply rwfp_cop; assumption.

              so (rwfs_cop_group _#4 Htrco2 (rwfs_app_incl _#4 (rwfs_app_incl _#4 Hrwfsd2)) Hcop21d) as [(_ & Hrwfsd) | Hcopd2].
                so (star_plus _#4 Hrwfsd) as [Heq | Hrwfdd].
                  subst; destruct Hneqd; auto; done.
    
                  destruct Hlast1.
                  destruct Hrwfdd as (i3d & Hrwf3 & _).
                  exists i3d.
                  eauto using rwf_truncate; done.
                
                (* The heart of the proof, such as it is, is here. *)
                exfalso.
                destruct Hcoacy2; [].
                exists i1d.
                eapply plus_trans; eauto; done.

      clear Hrelprop.
      exists (H2; ev_co i2 i1).
      autorewrite with canonlist.
      assert (trco (H;; H1;; H2; ev_co i2 i1)) as Htrco2'.
        eapply trco_co; eauto using writesto_app_incl; done.
      do2 3 split.
        assumption.
  
        eapply Hprop; eauto; done.
  
        intros i1' Hin.
        destruct Hin as [Heq | Hin].
          subst i1'.
          right; right.
          apply plus_one; [].
          apply co_fiat; left; reflexivity.
  
          so (Htotal2 _ Hin) as [Hcop | [Heq | Hcop]]; auto using cop_cons_incl; done.
          
        apply Forall_cons; eauto; done.

    exists (H2; ev_co i1 i2).
    autorewrite with canonlist.
    assert (trco (H;; H1;; H2; ev_co i1 i2)) as Htrco2'.
      eapply trco_co; eauto using writesto_app_incl; done.
    do2 3 split.
      assumption.

      eapply Hprop; eauto; done.

      intros i2' Hin.
      destruct Hin as [Heq | Hin].
        subst i2'.
        left.
        apply plus_one; [].
        apply co_fiat; left; reflexivity.

        so (Htotal2 _ Hin) as [Hcop | [Heq | Hcop]]; auto using cop_cons_incl; done.
        
      apply Forall_cons; eauto; done.
Qed.          


Lemma complete_co :
  forall H,
    trco H
    -> acyclic (co H)
    -> exists H' H'',
         H' = H;; H''
         /\ trco H'
         /\ acyclic (co H')
         /\ (forall i i' l, writesto H' i l -> writesto H' i' l -> cop H' i i' \/ i = i' \/ cop H' i' i)
         /\ Forall (fun h => exists i i', h = ev_co i i') H''.
Proof.
intros H Htrco Hcoacy.
so (locations_itemizable H) as (L & Hiniff).
cut (exists H' H'',
       H' = H;; H''
       /\ trco H'
       /\ acyclic (co H')
       /\ (forall i i' l, In l L -> writesto H' i l -> writesto H' i' l -> cop H' i i' \/ i = i' \/ cop H' i' i)
       /\ Forall (fun h => exists i i', h = ev_co i i') H'').
  intros (H' & H'' & Heq & Htrco' & Hcoacy' & Htotal & Hallco).
  exists H', H''.
  do2 4 split; auto; [].
  intros i1 i2 l Hwrite1 Hwrite2.
  eapply Htotal; eauto; [].
  apply Hiniff; [].
  exists i1.
  subst H'.
  eapply writesto_truncate; eauto; [].
  intros (i & a & His).
  so (Forall_forall _#3 andel Hallco _ His) as Heq.
  simpl in Heq; destruct_all Heq; discriminate Heq.
clear Hiniff.
induction L as [| l L IH].

(* nil *)
exists H, nil.
repeat2 split; auto; [].
intros ? ? ? Hin.
destruct Hin; done.

(* cons *)
destruct IH as (? & H1 & -> & Htrco1 & Hcoacy1 & Htotal1 & Hallco1).
so (complete_co_l _ l Htrco1 Hcoacy1) as (? & H2 & -> & Htrco2 & Hcoacy2 & Htotal2 & Hallco2).
exists (H;; H1;; H2), (H1;; H2).
autorewrite with canonlist.
#2 repeat2 split; auto.
  intros i1 i2 l' Hin Hwrite1 Hwrite2.
  destruct Hin as [Heq | Hin].
    subst l'.
    apply Htotal2; auto; done.

    assert (~ H2 {{ ev_is i a | i a }}) as Hnis.
      intros (? & ? & His).
      so (Forall_forall _#3 andel Hallco2 _ His) as Heq.
      simpl in Heq; destruct_all Heq; discriminate Heq.
    exploit (Htotal1 i1 i2 l') as [Hcop | [Heq | Hcop]]; eauto using writesto_truncate, cop_app_incl; done.
    
  apply Forall_app; auto; done.
Qed.


Lemma complete_history :
  forall H,
    consistent H
    -> exists H' H'',
         H' = H;; H''
         /\ trco H'
         /\ allexec H'
         /\ acyclic (co H')
         /\ (forall i i' l, writesto H' i l -> writesto H' i' l -> cop H' i i' \/ i = i' \/ cop H' i' i)
         /\ ~ H'' {{ ev_is i a | i a }}.
Proof.
intros H Hcons.
destruct Hcons as (H1 & Htrco & Hcoacy & Hallexec & Hnis1).
so (complete_co _ Htrco Hcoacy) as (? & H2 & -> & Htrco' & Hcoacy' & Htotal & Hallco).     
exists (H;; H1;; H2), (H1;; H2).
autorewrite with canonlist.
assert (~ H2 {{ ev_is i a | i a }}) as Hnis2.
  intros (? & ? & His).
  so (Forall_forall _#3 andel Hallco _ His) as Heq.
  simpl in Heq; destruct_all Heq; discriminate Heq.
#2 repeat2 split; auto.
  intros i p Hinit.
  so (in_app_or _#4 Hinit) as [Hinit' | Hinit'].
    so (Forall_forall _#3 andel Hallco _ Hinit') as Heq.
    simpl in Heq; destruct_all Heq; discriminate Heq.

    apply exec_app_incl; [].
    eapply Hallexec; eauto; done.

  intros (i & a & His).
  so (in_app_or _#4 His) as [His' | His'].
    destruct Hnis2; eauto; done.

    destruct Hnis1; eauto; done.
Qed.


Lemma sequentializable_anti :
  forall R R' H,
    (forall i i', R' i i' -> R i i')
    -> sequentializable R H
    -> sequentializable R' H.
Proof.
intros R R' H Himpl Hseq.
destruct Hseq as (Q & Hirrefl & Htrans & Htotal & Hresp & Hrfcon).
exists Q.
do2 4 split; auto; done.
Qed.


Lemma truncate_sequentializable :
  forall R H H',
    sequentializable R (H;; H')
    -> sequentializable R H.
Proof.
intros R H H' Hseq.
destruct Hseq as (Q & Hirrefl & Htrans & Htotal & Hresp & Hrfcon).
exists Q.
#2 do2 4 split; auto.
  intros i i' (p & Hinit) (p' & Hinit').
  apply Htotal; eexists; apply in_or_app; right; eauto; done.

  intros iw ir Hrf.
  #2 exploit (Hrfcon iw ir) as (Hq & Hnomid).
    apply in_or_app; right; assumption.

    split; auto; [].
    contradict Hnomid.
    destruct Hnomid as (l & iw' & Hreads & Hwritesto & Hq1 & Hq2).
    exists l, iw'.
    do2 3 split; auto using reads_app_incl, writesto_app_incl; done.
Qed.


Lemma extend_racefree :
  forall F H H',
    trco (H;; H')
    -> ~ H' {{ ev_is i a | i a }}
    -> racefree F H
    -> racefree F (H;; H').
Proof.
intros F H H' Htrco Hnis Hracefree.
unfold racefree.
intros i1 i2 p1 p2 Hinit1 Hinit2 Hneq Hconflict Hf.
#5 exploit (Hracefree i1 i2 p1 p2) as [Hsb | Hsb]; auto.
  so (conflict_is _#3 Hconflict andel) as (a & His).
  so (in_app_or _#4 His) as [His' | His'].
    destruct Hnis; eauto; done.

    so (is_inited _#3 (trco_truncate _ _ Htrco) His') as (p & Hinit).
    replace p with p1 in Hinit; auto; [].
    eapply (init_unique (H;; H')); eauto; [].
    apply in_or_app; right; assumption.
      
  so (conflict_is _#3 Hconflict ander) as (a & His).
  so (in_app_or _#4 His) as [His' | His'].
    destruct Hnis; eauto; done.

    so (is_inited _#3 (trco_truncate _ _ Htrco) His') as (p & Hinit).
    replace p with p2 in Hinit; auto; [].
    eapply (init_unique (H;; H')); eauto; [].
    apply in_or_app; right; assumption.

  eapply conflict_truncate; eauto; done.

  left; apply sb_app_incl; auto; done.

  right; apply sb_app_incl; auto; done.
Qed.


Corollary sequential_consistency_1 :
  forall H,
    consistent H
    -> sequentializable (ppo H) H.
Proof.
intros H Hconsistent.
so (complete_history _ Hconsistent) as (H' & H'' & Heq & Htrco & Hallexec & Hcoacy & Hcototal & _).
thus (acyclic (sc1 H')) by sc1_acyclic.
assert (sequentializable (ppo H') H').
  #2 eapply sequential_consistency; eauto.
    intros.
    apply ppo_decidable; auto; done.

    exists (ids H').
    intros i i' Hppo.
    destruct Hppo as (? & _ & Hvop & Hxop).
    #2 split; apply init_ids.
      eapply vop_inited_fst; eauto; done.

      eapply xop_inited_snd; eauto; done.
#2 apply (sequentializable_anti (ppo H')).
  intros i i' Hppo.
  subst.
  apply ppo_app_incl; auto; done.

  subst.
  eapply truncate_sequentializable; eauto; done.
Qed.


Corollary sequential_consistency_2 :
  forall H,
    consistent H
    -> racefree (fun _ => True) H
    -> sequentializable (po H) H.
Proof.
intros H Hconsistent Hracefree.
so (complete_history _ Hconsistent) as (H' & H'' & Heq & Htrco & Hallexec & Hcoacy & Hcototal & Hnis).
assert (racefree (fun _ => True) H') as Hracefree'.
  subst.
  apply extend_racefree; auto; done.
thus (acyclic (sc2 H')) by sc2_acyclic.
assert (sequentializable (po H') H').
  #2 eapply sequential_consistency; eauto.
    intros; apply po_decidable; done.

    exists (ids H').
    intros.
    apply po_ids; assumption.
#2 apply (sequentializable_anti (po H')).
  intros i i' Hpo.
  subst.
  apply po_app_incl; auto; done.

  subst.
  eapply truncate_sequentializable; eauto; done.
Qed.


Corollary sequential_consistency_3 :
  forall F H,
    (forall i, decidable (F i))
    -> consistent H
    -> racefree F H
    -> (forall i i' l, accesses H i l -> accesses H i' l -> F i -> F i')
    -> sequentializable (union (sppo H) (fpo F H)) H.
Proof.
intros F H Hdec Hconsistent Hracefree Hdisjoint.
so (complete_history _ Hconsistent) as (H' & H'' & Heq & Htrco & Hallexec & Hcoacy & Hcototal & Hnis).
assert (racefree F H') as Hracefree'.
  subst.
  apply extend_racefree; auto; done.
assert (acyclic (sc3 F H')).
  eapply sc3_acyclic; eauto; [].
  intros i i' l Hacc Hacc' Hf.
  subst.
  eapply Hdisjoint; eauto using accesses_truncate; done.
assert (sequentializable (union (sppo H') (fpo F H')) H').
  #2 eapply sequential_consistency; eauto.
    intros i i'.
    #2 apply dec_or.
      apply sppo_decidable; auto; done.

      #2 apply dec_and.
        apply Hdec; done.
        
        apply po_decidable; done.

    exists (ids H').
    intros i i'.
    intros [Hsppo | Hfpo].
      destruct Hsppo as (? & _ & _ & Hvop & Hxop).
      #2 split; apply init_ids.
        eapply vop_inited_fst; eauto; done.

        eapply xop_inited_snd; eauto; done.

      destruct Hfpo as (? & Hpo).
      apply po_ids; assumption.
#2 apply (sequentializable_anti (union (sppo H') (fpo F H'))).
  intros i i' [Hppo | Hfpo].
    left.
    subst; apply sppo_app_incl; auto; done.

    right.
    subst; apply fpo_app_incl; auto; done.

  subst.
  eapply truncate_sequentializable; eauto; done.
Qed.
