
Require Import TraverseTactics.
Require Export PreSyntax.

Require Import Tactics.
Require Import List.


Inductive attr : Set :=
| at_x : attr
| at_v : attr.


Inductive term : Set :=
| var : nat -> term

(* tags *)

| tg_lit : tag -> term

(* types *)

| tp_unit : term
| tp_base : term
| tp_susp : term -> term
| tp_ref : term -> term
| tp_arrow : term -> term -> term

(* terms *)

| tm_unit : term

| tm_lit : base -> term
| tm_prim : (* tm *) term -> (* tm *) term -> term
| tm_if : (* tm *) term -> (* tm *) term -> (* tm *) term -> term

| tm_loc : location -> term

| tm_susp : (* xp *) term -> term

| tm_lam : (* tp *) term -> (* tm => *) term -> term
| tm_fun : (* tp *) term -> (* tp *) term -> (* tm => tm => *) term -> term
| tm_app : term -> term -> term

(* fences *)

| fc_tag : (* tg *) term -> term
| fc_pre : attr -> term
| fc_post : attr -> term

(* fence lists *)

| fl_nil : term
| fl_cons : (* fc *) term -> term -> term

(* expressions *)

| xp_ret : (* tm *) term -> term
| xp_force : (* tm *) term -> term
| xp_bind : (* tp *) term -> term -> (* tm => *) term -> term
| xp_spec : (* tm *) term -> (* tm *) term -> term -> term

| xp_read : (* tm *) term -> term
| xp_write : (* tm *) term -> (* tm *) term -> term
| xp_rw : (* tm *) term -> (* tm *) term -> term
| xp_rmw :  (* tp *) term -> (* tm *) term -> (* tm => tm *) term -> term
| xp_push : term
| xp_nop : term

| xp_action : identifier -> term

| xp_new : attr -> (* tg => tg => *) term -> term
| xp_fence : (* fc *) term -> term -> term

(* actions *)

| ac_read : location -> term
| ac_write : location -> (* tm *) term -> term
| ac_rw : location -> (* tm *) term -> term
| ac_push : term
| ac_nop : term

(* transactions *)

| tr_nothing : term
| tr_init : (* fl *) term -> identifier -> (* ac *) term -> term
| tr_exec : identifier -> (* tm *) term -> term
| tr_new : attr -> tag -> tag -> term

(* classifiers *)

| cl_vl : (* tp *) term -> term
| cl_tg : term

(* nonsense *)

| nonsense : term
.


Inductive event : Set :=
| ev_init : identifier -> thread -> event
| ev_is : identifier -> (* ac *) term -> event
| ev_exec : identifier -> event
| ev_rf : identifier -> identifier -> event
| ev_edge : attr -> tag -> tag -> event
| ev_fence : (* fc *) term -> identifier -> event
| ev_co : identifier -> identifier -> event
.


Definition xstate := list (thread * (* xp *) term).
Definition history := list event.

Definition tgtp := list tag.
Definition lctp := list (location * (* tp *) term).
Definition idtp := list (identifier * ((* tp *) term * thread)).

Definition state := (tgtp * lctp * idtp * history * xstate)%type.



Fixpoint traverse (S:Set) (enter : S -> S) (resolve : S -> nat -> term) (s : S) (t : term) {struct t} : term :=
  (match t with
   | var i => resolve s i

   | tg_lit tg => tg_lit tg

   | tp_unit => tp_unit
   | tp_base => tp_base
   | tp_susp t1 =>
       tp_susp (traverse S enter resolve s t1)
   | tp_ref t1 =>
       tp_ref (traverse S enter resolve s t1)
   | tp_arrow t1 t2 =>
       tp_arrow (traverse S enter resolve s t1) (traverse S enter resolve s t2)

   | tm_unit => tm_unit
   | tm_lit c => tm_lit c
   | tm_prim t1 t2 =>
       tm_prim (traverse S enter resolve s t1) (traverse S enter resolve s t2) 
   | tm_if t1 t2 t3 =>
       tm_if (traverse S enter resolve s t1) (traverse S enter resolve s t2) (traverse S enter resolve s t3)
   | tm_loc l => tm_loc l
   | tm_susp t1 =>
       tm_susp (traverse S enter resolve s t1)
   | tm_lam t1 t2 =>
       tm_lam (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | tm_fun t1 t2 t3 =>
       tm_fun (traverse S enter resolve s t1) (traverse S enter resolve s t2) (traverse S enter resolve (enter (enter s)) t3)
   | tm_app t1 t2 =>
       tm_app (traverse S enter resolve s t1) (traverse S enter resolve s t2) 

   | fc_tag t1 =>
       fc_tag (traverse S enter resolve s t1)
   | fc_pre b => fc_pre b
   | fc_post b => fc_post b

   | fl_nil => fl_nil
   | fl_cons t1 t2 =>
       fl_cons (traverse S enter resolve s t1) (traverse S enter resolve s t2)

   | xp_ret t1 =>
       xp_ret (traverse S enter resolve s t1)
   | xp_force t1 =>
       xp_force (traverse S enter resolve s t1)
   | xp_bind t1 t2 t3 =>
       xp_bind (traverse S enter resolve s t1) (traverse S enter resolve s t2) (traverse S enter resolve (enter s) t3)
   | xp_spec t1 t2 t3 =>
       xp_spec (traverse S enter resolve s t1) (traverse S enter resolve s t2) (traverse S enter resolve s t3)
   | xp_read t1 => xp_read (traverse S enter resolve s t1)
   | xp_write t1 t2 =>
       xp_write (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | xp_rw t1 t2 =>
       xp_rw (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | xp_rmw t1 t2 t3 =>
       xp_rmw (traverse S enter resolve s t1) (traverse S enter resolve s t2) (traverse S enter resolve (enter s) t3)
   | xp_push => xp_push
   | xp_nop => xp_nop
   | xp_action i => xp_action i
   | xp_new b t1 =>
       xp_new b (traverse S enter resolve (enter (enter s)) t1)
   | xp_fence t1 t2 =>
       xp_fence (traverse S enter resolve s t1) (traverse S enter resolve s t2)

   | ac_read l => ac_read l
   | ac_write l t1 =>
       ac_write l (traverse S enter resolve s t1)
   | ac_rw l t1 =>
       ac_rw l (traverse S enter resolve s t1)
   | ac_push => ac_push
   | ac_nop => ac_nop

   | tr_nothing => tr_nothing
   | tr_init t1 i t2 =>
       tr_init (traverse S enter resolve s t1) i (traverse S enter resolve s t2)
   | tr_exec i t1 =>
       tr_exec i (traverse S enter resolve s t1)
   | tr_new b tag1 tag2 => tr_new b tag1 tag2

   | cl_vl t1 =>
       cl_vl (traverse S enter resolve s t1)
   | cl_tg => cl_tg

   | nonsense => nonsense
   end).


Lemma traverse_var :
  forall S enter resolve s i,
    traverse S enter resolve s (var i) = resolve s i.
Proof.
auto.
Qed.


Lemma traverse_parametric :
  forall (S:Set) (S':Set) (R : S -> S' -> Prop) enter enter' resolve resolve' s s' t,
    (forall s s', R s s' -> R (enter s) (enter' s'))
    -> (forall s s' i, R s s' -> resolve s i = resolve' s' i)
    -> R s s'
    -> traverse S enter resolve s t = traverse S' enter' resolve' s' t.
Proof.
prove_traverse_parametric term_ind.
Qed.


Lemma traverse_id :
  forall (S:Set) (R : S -> Prop) enter resolve s t,
    (forall s, R s -> R (enter s))
    -> (forall s i, R s -> resolve s i = var i)
    -> R s
    -> traverse S enter resolve s t = t.
Proof.
prove_traverse_id term_ind.
Qed.


Lemma traverse_compose :
  forall (S:Set) (S':Set) enter enter' resolve resolve' s s' t,
    traverse S enter resolve s (traverse S' enter' resolve' s' t)
    = traverse (S * S')
      (fun p => let (s, s') := p in (enter s, enter' s'))
      (fun p i => let (s, s') := p in traverse S enter resolve s (resolve' s' i))
      (s, s') t.
Proof.
prove_traverse_compose term_ind.
Qed.


Definition termx := term.
Definition varx := var.
